# Api Entities (Dto)
from .version import __version__
from .api_dto import ApiDto, VarType
from .paged_query_result import PagedQueryResult
from .plot import Plot
from .mlmodel import ModelInfo, ModelList, MLModelConfig, ModelFile, ModelIdentifierInfo
from .request import Request, filter_map, RequestGroup, RequestGroupMap, DynamicSelector
from .execution import Execution, ExecutionStatus, ExecutionStepLog, AbortedException
from .experiment import Experiment
from .ds_dataframe import DSDataFrame
from .script import Script, ScriptConfig
from .template import Template, TemplateProperty
from .solution_component import SolutionComponent, SolutionType
from .business_label import BusinessLabel
from .twinregistration import TwinRegistration, TwinRegistrationProperty
from .trigger import Trigger
from .group_system import GroupSystem
from .bucket import Bucket

# Sql Entities (Dto)
from .twin import Twin, TwinBlockType, TwinType
from .datapoint import DataPoint, BusinessType, Label, Unit, Category, InputModeType
from .datastore import DataStore
from .insight import Insight

# Api
from .api_interface import ApiInterface
from .api_config import _registry
from .wizata_dsapi_client import api
from .wizata_dsapi_client import WizataDSAPIClient
from .dataframe_toolkit import df_to_json, df_to_csv, df_from_json, df_from_csv, validate, generate_epoch, \
    verify_relative_datetime, generate_unique_key, df_to_dict, df_from_dict
from .model_toolkit import predict

# Legacy
from .dsapi_json_encoder import DSAPIEncoder
from .wizard_function import WizardStep, WizardFunction
from .wizard_request import WizardRequest

# Pipeline Entities (Dto)
from .pipeline import Pipeline, PipelineStep, StepType, WriteConfig, PipelineIO, RunnerStatus, AlertType
from .context import Context
from .ilogger import ILogger
from .pipeline_image import PipelineImage
from .evaluation import Evaluation

# Streamlit utils
from .streamlit_utils import (get_streamlit_token, get_streamlit_domain, get_streamlit_from, get_streamlit_to,
                              get_streamlit_twin_id)



from .api_dto import ApiDto


class Bucket(ApiDto):
    """
    a bucket is a group of time-series data.
    """

    @classmethod
    def route(cls):
        return "buckets"

    @classmethod
    def from_dict(cls, data):
        obj = Bucket()
        obj.from_json(data)
        return obj

    def __init__(self,
                 name: str = None,
                 retention_seconds: int = None):
        self.name = name
        self.retention_seconds = retention_seconds

    def api_id(self) -> str:
        return str(self.name).upper()

    def endpoint(self) -> str:
        return "Buckets"

    def to_json(self, target: str = None):
        obj = {}
        if self.retention_seconds is not None:
            obj["retentionSeconds"] = int(self.retention_seconds)
        if self.name is not None:
            obj["name"] = str(self.name)
        return obj

    def from_json(self, obj):
        if "retentionSeconds" in obj.keys() and obj["retentionSeconds"] is not None:
            self.retention_seconds = int(obj["retentionSeconds"])
        if "name" in obj.keys() and obj["name"] is not None:
            self.name = obj["name"]



animals = [
    "Antelope", "Armadillo", "Alligator", "Cheetah", "Dolphin", "Elephant", "Firefly", "Flamingo", "Giraffe", "Hamster",
    "Kangaroo", "Koala", "Lemming", "Manatee", "Meerkat", "Mongoose", "Mosquito", "Ocelot", "Penguin", "Raccoon",
    "Salamander", "Starfish", "Termite", "Wallaby", "Walrus", "Weasel", "Wolverine", "Woodlouse", "Woodpecker", "Zebra",
    "Cat", "Dog", "Bat", "Fox", "Cow", "Pig", "Rat", "Mole", "Goat", "Fish",
    "Bird", "Frog", "Toad", "Wolf", "Bear", "Deer", "Lion", "Horse", "Mule", "Moth",
    "Mouse", "Sheep", "Snake", "Whale", "Shark", "Hawk", "Eagle", "Owl", "Crow", "Duck",
    "Swan", "Goose", "Puma", "Civet", "Seal", "Crab", "Fossa", "Hyena", "Tiger", "Jaguar",
    "Leopard", "Lizard", "Gibbon", "Gorilla", "Baboon", "Monkey", "Koala", "Panda", "Toucan", "Kiwi"
]

colors = [
    "Amaranth", "Amber", "Amethyst", "Apricot", "Aquamarine", "Azure", "Beige", "Black", "Blue", "Bronze",
    "Brown", "Carmine", "Cerise", "Cobalt", "Copper", "Coral", "Crimson", "Cyan", "Emerald", "Fuchsia",
    "Gold", "Gray", "Green", "Indigo", "Ivory", "Lavender", "Lemon", "Lilac", "Magenta", "Maroon",
    "Mauve", "Olive", "Orange", "Orchid", "Peach", "Periwinkle", "Pink", "Plum", "Purple", "Red",
    "Ruby", "Salmon", "Sapphire", "Scarlet", "Silver", "Tan", "Teal", "Turquoise", "Violet", "White", "Yellow",
    "Aqua", "Beige", "Coral", "Cyan", "Green", "Khaki", "Lemon", "Lilac", "Peach", "Plum", 
    "Salmon", "Teal", "Violet", "Wheat", "Bronze", "Copper", "Gold", "Silver", "Azure", "Pearl"
]



import uuid
from enum import Enum


class ApiDtoInterface:

    def load_model(self, model):
        pass


class VarType(Enum):
    """
    defines possible type for a defined variable.
        - "string" text sequence of characters
        - "integer" signed 64-bit integer
        - "datetime" static datetime or timestamp (UTC)
        - "relative" relative datetime using "now+-?yMwdmsms" format (e.g. 'now-7d')
        - "datapoint" hardware id, uuid or wizata_dsapi.DataPoint defining a datapoint
        - "float" floating-point decimal number
        - "json" complex object defined as a serializable JSON document.
    """
    STRING = "string"
    INTEGER = "integer"
    DATETIME = "datetime"
    RELATIVE = "relative"
    DATAPOINT = "datapoint"
    FLOAT = "float"
    JSON = "json"


class ApiDto:

    def api_id(self) -> str:
        """
        return current object id on Web API format.
        """
        pass

    def endpoint(self) -> str:
        """
        return endpoint name used to contact backend.
        """
        pass

    def to_json(self, target: str = None):
        """
        transform current object into a dict that could be JSONIFY.
        :target: by default - None. Can be 'backend-create' or 'backend-update' in order to JSONIFY only backend fields.
        :return: dumpable dict.
        """
        pass

    def from_json(self, obj):
        """
        load the object from a dict originating of a JSON format.
        :param obj: object to load information from.
        """
        pass

    def set_id(self, id_value):
        """
        specify the id_value neutrally
        :param id_value:
        :return:
        """
        pass

    @classmethod
    def route(cls):
        """
        Endpoint name in Web API (v0.4+).
        """
        pass

    @classmethod
    def from_dict(cls, data):
        """
        Init object instance from dict (v0.4+)
        """
        pass

    @classmethod
    def get_type(cls):
        """
        Return type of get format json by default - override with either pickle, dill, ...
        """
        return "json"

    @classmethod
    def get_id_type(cls) -> type:
        """
        return type of the id format, by default UUID but some are overridden integer.
        """
        return uuid.UUID



import wizata_dsapi

import pandas
import numpy

import sklearn
import sklearn.cluster
import sklearn.metrics
import sklearn.ensemble


def filter_df(context: wizata_dsapi.Context):
    """
    filter a dataframe
    :param context:
    :return:
    """

    if "filters" not in context.properties or not isinstance(context.properties['filters'], list):
        raise ValueError(f'there is no list *filters* in properties - please set them on context or config')

    df = context.dataframe.copy()

    filters = context.properties['filters']
    for filter_row in filters:
        try:
            df = df.query(filter_row)
        except pandas.errors.ParserError as e:
            raise ValueError(f"error parsing filter string '{filter_row}': {e}")

    return df


def clustering(context: wizata_dsapi.Context):
    """
    clustering
    :param context:
    :return:
    """
    df = context.dataframe.copy()
    scaler = sklearn.preprocessing.StandardScaler()
    df_clustering_scaler = scaler.fit_transform(df)

    range_n_clusters = list(range(2, min(10, df_clustering_scaler.shape[0])))
    silhouette_avg = []
    for num_clusters in range_n_clusters:
        kmeans = sklearn.cluster.KMeans(n_clusters=num_clusters)
        kmeans.fit(df_clustering_scaler)
        cluster_labels = kmeans.labels_
        unique, counts = numpy.unique(cluster_labels, return_counts=True)

        if len(unique) >= 2:
            silhouette_avg.append(sklearn.metrics.silhouette_score(df_clustering_scaler, cluster_labels))
        else:
            silhouette_avg.append(numpy.nan)

    if numpy.isnan(silhouette_avg).all():
        df['cluster_labels'] = 0
    else:
        best_nb_clusters = silhouette_avg.index(max(silhouette_avg)) + 2
        kmeans = sklearn.cluster.KMeans(n_clusters=best_nb_clusters)
        kmeans.fit(df_clustering_scaler)
        cluster_labels = kmeans.labels_
        df['cluster_labels'] = cluster_labels
        df['cluster_labels'] = df['cluster_labels'].apply(lambda x: int(x + 1))

    return df


def merge(context: wizata_dsapi.Context):
    """
    merge
    :param context:
    :return:
    """
    dataframes = context.current_dataframes()
    if len(dataframes) <= 1:
        raise ValueError(f'there is not enough dataframes to concat')

    how = "outer"
    if "how" in context.properties:
        how = context.properties["how"]

    df = None
    for key in dataframes:
        if df is None:
            df = dataframes[key]
        else:
            df = df.merge(dataframes[key], how=how, left_index=True, right_index=True)
    return df


def fillna(context: wizata_dsapi.Context):
    """
    fillna
    :param context:
    :return:
    """
    df = context.dataframe

    if "fillna" not in context.properties:
        raise KeyError(f'please set a property dict fillna')

    for key in context.properties["fillna"]:
        df[key] = df[key].fillna(value=context.properties["fillna"][key])

    return df


def target_feat_to_binary(context: wizata_dsapi.Context):
    """
    target_feat_to_binary
    :param context:
    :return:
    """
    df = context.dataframe

    if "target_feat" not in context.properties:
        raise KeyError(f'please set a target feature to transform to binary class')

    target_feat = context.properties["target_feat"]["sensor"]
    operator = context.properties["target_feat"]["operator"]
    threshold = context.properties["target_feat"]["threshold"]

    if operator == 'lt':
        df[target_feat] = numpy.where(df[target_feat] < threshold, 1, 0)
    elif operator == 'lte':
        df[target_feat] = numpy.where(df[target_feat] <= threshold, 1, 0)
    elif operator == 'gt':
        df[target_feat] = numpy.where(df[target_feat] > threshold, 1, 0)
    elif operator == 'gte':
        df[target_feat] = numpy.where(df[target_feat] >= threshold, 1, 0)
    else:
        raise KeyError(f'operator type for binarisation not know')

    # Check if at least 1 value of each class
    if df[target_feat].nunique() == 1:
        raise KeyError(f'classification model requires 2 classes, only one was detected')
    elif df[target_feat].nunique() > 2:
        raise KeyError(f'classification model requires 2 classes, more than 2 were detected')

    return df



import uuid
import json
import sys

import pandas
import wizata_dsapi

from .api_dto import ApiDto, VarType
from .script import ScriptConfig, Script
from .request import Request
from .mlmodel import MLModelConfig
from .template import Template, TemplateProperty
from .twin import Twin
from .twinregistration import TwinRegistration, TwinRegistrationProperty

from enum import Enum


class RunnerStatus(Enum):
    """
    defines a runner status.
        - starting: runner is starting but not yet listening to the queue
        - listening: runner is listening to the queue
        - executing: runner is executing some package and will continue listening
        - pausing: runner is executing some pipeline and will stop listening
        - idle: runner is stopped and is doing nothing
        - terminating: runner is executing and will terminate the process (and then restart auto)
    """
    STARTING = 'starting'
    LISTENING = 'listening'
    EXECUTING = 'executing'
    PAUSING = 'pausing'
    IDLE = 'idle'
    TERMINATING = 'terminating'


class StepType(Enum):
    """
    defines a pipeline step type.
        - "query"
        - "script"
        - "model"
        - "write"
        - "plot"
        - "df_save"
    """
    QUERY = 'query'
    SCRIPT = 'script'
    MODEL = 'model'
    WRITE = 'write'
    PLOT = 'plot'
    DF_SAVE = 'df_save'


class AlertType(Enum):
    """
    defines a type of alert.
        - "sms"
        - "email"
    """
    SMS = 'sms'
    EMAIL = 'email'
    SLACK = 'slack'
    TEAMS = 'teams'
    WHATSAPP = 'whatsapp'


class WriteConfig:
    """
    define how a step should write data into Wizata.

    can use a mapping table (key=dataframe column, value=hardware id or template property).

    :ivar uuid.UUID config_id: technical id of this configuration.
    :ivar dict datapoints: mapping dict between dataframe column and Wizata datapoint hardware id or template property.
    :ivar str map_property_name: define name of a context property containing a mapping table to use.
    """

    def __init__(self,
                 config_id = None,
                 datapoints: dict = None,
                 map_property_name: str = None,
                 bucket: str = None,
                 writer_type: str = None):
        if config_id is None:
            config_id = uuid.uuid4()
        self.config_id = config_id
        self.datapoints = datapoints
        self.map_property_name = map_property_name
        self.bucket = bucket
        self.writer_type = writer_type

    def from_json(self, obj):
        if "id" in obj:
            self.config_id = obj["id"]
        if "datapoints" in obj:
            self.datapoints = obj["datapoints"]
        if "map_property_name" in obj and obj["map_property_name"] is not None:
            self.map_property_name = str(obj["map_property_name"])
        if "bucket" in obj and obj["bucket"] is not None:
            self.bucket = obj["bucket"]
        if "writerType" in obj and obj["writerType"] is not None:
            self.writer_type = obj["writerType"]

    def to_json(self):
        obj = {
            "id": str(self.config_id)
        }
        if self.datapoints is not None:
            obj["datapoints"] = self.datapoints
        if self.bucket is not None:
            obj["bucket"] = self.bucket
        if self.map_property_name is not None:
            obj["map_property_name"] = self.map_property_name
        if self.writer_type is not None:
            obj["writerType"] = self.writer_type
        return obj


class PipelineIO(ApiDto):
    """
    define an input or output of a pipeline step.
    :ivar str dataframe: name of the dataframe to connect input and output, used in context.dafaframes
    :ivar dict mapping: mapping to rename dataframe column from key to value.
    :ivar list drops: list of columns to drop from dataframe.
    :ivar list columns: list of columns to select from dataframe and exclude others.
    """

    def __init__(self, dataframe: str = None, mapping: dict = None, drops: list = None, columns: list = None):
        self.dataframe = dataframe
        self.mapping = mapping
        self.drops = drops
        self.columns = columns

    def from_json(self, obj):
        if "dataframe" in obj:
            self.dataframe = obj["dataframe"]
        else:
            raise ValueError(f'a pipeline I/O should have at least a dataframe name.')
        if "mapping" in obj and obj["mapping"] is not None:
            if not isinstance(obj["mapping"], dict):
                raise TypeError(f'mapping should be dict not a {obj["mapping"].__class__.__name__}')
            self.mapping = obj["mapping"]
        if "drops" in obj and isinstance(obj['drops'], list):
            self.drops = obj['drops']
        if "columns" in obj and isinstance(obj['columns'], list):
            self.columns = obj['columns']

    def to_json(self, target: str = None):
        if self.dataframe is None or not isinstance(self.dataframe, str):
            raise ValueError(f'a pipeline I/O should have at least a dataframe name.')
        obj = {
            "dataframe": self.dataframe
        }
        if self.mapping is not None and isinstance(self.mapping, dict):
            obj["mapping"] = self.mapping
        if self.drops is not None and isinstance(self.drops, list):
            obj["drops"] = self.drops
        if self.columns is not None and isinstance(self.columns, list):
            obj["columns"] = self.columns
        return obj

    def _prepare_df(self, df: pandas.DataFrame) -> pandas.DataFrame:
        """
        prepare dataframe in both 3.9 and 3.11+
        :param pandas.DataFrame df: dataframe to prepare.
        :return: prepared dataframe.
        """

        try:
            prepare_df = df.copy()

            # map the dataframe
            if self.mapping is not None:
                for old_name, new_name in self.mapping.items():
                    if old_name not in prepare_df.columns:
                        raise ValueError(f"column '{old_name}' not found in the df.")
                    prepare_df = prepare_df.rename(columns={old_name: new_name})

            # select the column to keep or drop extra columns
            if self.columns is not None and len(self.columns) > 0:
                prepare_df = prepare_df[self.columns]
            elif self.drops is not None and len(self.drops) > 0:
                prepare_df = prepare_df.drop(columns=self.drops)

            return prepare_df
        except Exception as e:
            raise RuntimeError(f'not able to prepare your dataframe following Pipeline I/O directives {e}')

    if sys.version_info >= (3, 11):

        ## with torch support
        import torch
        from typing import Union

        def prepare(self, df: Union[pandas.DataFrame, torch.Tensor]) -> Union[pandas.DataFrame, torch.Tensor]:
            import torch
            if isinstance(df, torch.Tensor):
                return df
            else:
                return self._prepare_df(df)

    else:

        ## without torch support
        def prepare(self, df: pandas.DataFrame) -> pandas.DataFrame:
            return self._prepare_df(df)

    @classmethod
    def from_obj(cls, obj):
        """
        convert a str, dict or PipelineIO to a PipelineIO
        """
        if isinstance(obj, PipelineIO):
            return obj
        elif isinstance(obj, dict):
            pipeline_io = PipelineIO()
            pipeline_io.from_json(obj)
            return pipeline_io
        elif isinstance(obj, str):
            pipeline_io = PipelineIO(dataframe=obj)
            return pipeline_io
        else:
            raise TypeError(f'unsupported input type for a Pipeline I/O : {obj.__class__.__name__}')

    @classmethod
    def from_list(cls, obj_list: list) -> list:
        """
        get a list of Pipeline I/O from list containing Pipeline I/O, dict or str.
        :param obj_list: list to convert
        :return: list of Pipeline I/O
        """
        pipeline_io_list = []
        for obj in obj_list:
            pipeline_io_list.append(cls.from_obj(obj))
        return pipeline_io_list


class PipelineStep(ApiDto):
    """
    step of a pipeline.
    :ivar uuid.UUID step_id: id of pipeline step.
    :ivar wizata_dsapi.StepType step_type: step type.
    :ivar config: step configuration; request, script_config, model_config, ...
    :ivar inputs: list of wizata_dsapi.PipelineIO
    :ivar outputs: list of wizata_dsapi.PipelineIO
    """

    def __init__(self,
                 step_id: uuid.UUID = None,
                 step_type: StepType = None,
                 config=None,
                 inputs=None,
                 outputs=None):

        if outputs is None:
            outputs = []
        if inputs is None:
            inputs = []
        if step_id is None:
            step_id = uuid.uuid4()
        self.step_id = step_id
        self.step_type = step_type
        self.config = config
        self.inputs = inputs
        self.outputs = outputs

    def from_json(self, obj):
        if "id" in obj.keys():
            self.step_id = uuid.UUID(obj["id"])

        if "type" in obj.keys():
            self.step_type = StepType(obj["type"])
        else:
            raise TypeError(f'pipeline step must have a type.')
        if "config" in obj.keys():
            if self.step_type == StepType.SCRIPT:
                self.config = ScriptConfig()
                self.config.from_json(obj['config'])
            elif self.step_type == StepType.QUERY:
                self.config = Request()
                self.config.from_json(obj['config'])
            elif self.step_type == StepType.WRITE:
                self.config = WriteConfig()
                self.config.from_json(obj['config'])
            elif self.step_type == StepType.MODEL:
                self.config = MLModelConfig()
                self.config.from_json(obj['config'])
            elif self.step_type == StepType.PLOT:
                self.config = ScriptConfig()
                self.config.from_json(obj['config'])
            else:
                self.config = obj['config']
        self.inputs = []
        if "inputs" in obj.keys():
            self.inputs = PipelineIO.from_list(obj["inputs"])

        self.outputs = []
        if "outputs" in obj.keys():
            self.outputs = PipelineIO.from_list(obj["outputs"])

    def to_json(self, target: str = None):
        obj = {
            "id": str(self.step_id)
        }
        if self.step_type is not None:
            obj["type"] = str(self.step_type.value)
        if self.config is not None:
            if self.step_type == StepType.SCRIPT:
                obj["config"] = self.config.to_json()
            elif self.step_type == StepType.QUERY:
                obj["config"] = self.config.to_json()
            elif self.step_type == StepType.WRITE:
                obj["config"] = self.config.to_json()
            elif self.step_type == StepType.MODEL:
                obj["config"] = self.config.to_json()
            elif self.step_type == StepType.PLOT:
                obj["config"] = self.config.to_json()
            else:
                obj["config"] = self.config

        if self.inputs is not None:
            obj["inputs"] = []
            pipeline_io: PipelineIO
            for pipeline_io in self.get_inputs():
                obj["inputs"].append(pipeline_io.to_json())

        if self.outputs is not None:
            obj["outputs"] = []
            pipeline_io: PipelineIO
            for pipeline_io in self.get_outputs():
                obj["outputs"].append(pipeline_io.to_json())

        return obj

    def get_inputs(self) -> list:
        """
        get list of inputs, verified and transformed as PipelineIO
        :return: list of PipelineIO
        """
        return PipelineIO.from_list(self.inputs)

    def get_outputs(self) -> list:
        """
        get list of inputs, verified and transformed as PipelineIO
        :return: list of PipelineIO
        """
        return PipelineIO.from_list(self.outputs)

    def inputs_names(self, f_type: str = None) -> list:
        """
        get a list str representing inputs names
        :param f_type: filter on a type
        :return: list str
        """
        names = []

        if f_type is not None and f_type != "dataframe":
            raise ValueError(f'pipeline I/O only supports dataframe.')

        input_io: PipelineIO
        for input_io in self.get_inputs():
            names.append(input_io.dataframe)

        return names

    def outputs_names(self, f_type: str = None) -> list:
        """
        get a list str representing outputs names
        :param f_type: filter on a type
        :return: list str
        """
        names = []

        if f_type is not None and f_type != "dataframe":
            raise ValueError(f'pipeline I/O only supports dataframe.')

        output_io: PipelineIO
        for output_io in self.get_outputs():
            names.append(output_io.dataframe)

        return names

    def get_unique_input(self) -> PipelineIO:
        """
        verify pipeline step as only one input and returns it, raise an error otherwise.
        """
        inputs = self.get_inputs()
        if len(inputs) != 1:
            raise ValueError(f'pipeline step {self.step_id} of type {self.step_type.value} '
                             f'does not contains one and only one input but {len(inputs)}')
        return inputs[0]

    def get_unique_output(self) -> PipelineIO:
        """
        verify pipeline step as only one output and returns it, raise an error otherwise.
        """
        outputs = self.get_outputs()
        if len(outputs) != 1:
            raise ValueError(f'pipeline step {self.step_id} of type {self.step_type.value} '
                             f'does not contains one and only one output but {len(outputs)}')
        return outputs[0]

    def get_input(self, name: str) -> PipelineIO:
        """
        get input dict based on value name.
        :param name: value name to find.
        :return: input dict
        """
        input_io: PipelineIO
        for input_io in self.get_inputs():
            if name == input_io.dataframe:
                return input_io

    def get_output(self, name: str) -> PipelineIO:
        """
        get output dict based on value name.
        :param name: value name to find.
        :return: input dict
        """
        output_io: PipelineIO
        for output_io in self.get_outputs():
            if name == output_io.dataframe:
                return output_io


class Pipeline(ApiDto):
    """
    Pipeline defines a set of steps that can be executed together.
    :ivar uuid.UUID pipeline_id: technical id of the pipeline.
    :ivar uuid.UUID experiment_id: technical id of the experiment linked to the pipeline.
    :ivar str key: logical string unique id of the pipeline.
    :ivar list steps: list of steps connected between by their inputs and outputs.
    :ivar uuid.UUID template_id: template id associated to the pipeline.
    :ivar dict variables: dictionary of variable name with they wizata_dsapi.VarType.
    :ivar uuid.UUID createdById: unique identifier of creating user.
    :ivar int createdDate: timestamp of created date.
    :ivar uuid.UUID updatedById: unique identifier of updating user.
    :ivar int updatedDate: timestamp of updated date.
    """

    @classmethod
    def route(cls):
        return "pipelines"

    @classmethod
    def from_dict(cls, data):
        obj = Pipeline()
        obj.from_json(data)
        return obj

    def __init__(self,
                 pipeline_id: uuid.UUID = None,
                 key: str = None,
                 variables: dict = None,
                 steps: list = None,
                 plots: list = None,
                 template_id: uuid.UUID = None,
                 experiment_id: uuid.UUID = None):

        if pipeline_id is None:
            pipeline_id = uuid.uuid4()
        self.pipeline_id = pipeline_id
        self.key = key
        if variables is None:
            variables = {}
        self.variables = variables
        self.template_id = template_id
        self.experiment_id = experiment_id
        self.createdById = None
        self.createdDate = None
        self.updatedById = None
        self.updatedDate = None
        self.plots = plots
        if plots is None:
            self.plots = []
        if steps is not None:
            for step in steps:
                if not isinstance(step, PipelineStep):
                    raise TypeError(f'step expected PipelineStep but received {step.__class__.__name__}')
            self.steps = steps
        else:
            self.steps = []

        # related entities
        self.template = None
        self.twins = {}

    @property
    def key(self):
        return self._key

    @key.setter
    def key(self, value):
        if value is not None and len(value) > 32:
            raise ValueError(f'key is limited to 32 char : {value} ')
        self._key = value

    @key.deleter
    def key(self):
        del self._key

    def check_path(self) -> bool:
        """
        validate that steps create a valid path.
        return true if path is valid, otherwise raise errors
        """

        followed = []  # all steps already followed
        produced = []  # all outputs already produced

        steps_to_follow = self._next_steps(followed, produced)
        if len(steps_to_follow) == 0:
            raise ValueError('path does not contains any initial steps producing outputs')

        while len(steps_to_follow) > 0:
            for step in steps_to_follow:
                self._follow_step(step, followed, produced)
            steps_to_follow = []
            if len(followed) < len(self.steps):
                steps_to_follow = self._next_steps(followed, produced)

        if len(followed) == len(self.steps):
            return True
        else:
            raise RuntimeError(f'missing {len(self.steps)-len(followed)} step(s) that could not be followed')

    def _follow_step(self,
                     step: PipelineStep,
                     followed: list,
                     produced: list):
        """
        simulate that step have been followed.
        """
        if step in followed:
            raise RuntimeError(f'path cannot pass two times through same step {step.step_id}')
        followed.append(step)
        if step.step_type == StepType.QUERY and len(step.outputs) != 1:
            raise RuntimeError(f'query step must have exactly one output and be of type dataframe')
        if step.step_type == StepType.WRITE and (len(step.inputs) != 1 or len(step.outputs) > 0):
            raise RuntimeError(f'write step must have exactly one input and no outputs')
        for output in step.outputs_names():
            if output in produced:
                raise RuntimeError(f'output {output} is already produced.')
            produced.append(output)

    def _next_steps(self, followed, produced):
        """
        find all next steps that are ready to be executed
        """
        next_steps = []
        step: PipelineStep
        for step in self.steps:
            if all(s_input in produced for s_input in step.inputs_names()) and step not in followed:
                next_steps.append(step)
        return next_steps

    def add_query(self,
                  request: Request,
                  df_name: str = "query_df",
                  use_template: bool = True,
                  output_df: PipelineIO = None):
        """
        add a query step
        :param request: request definition to add.
        :param df_name: output name ot use for the dataframe - use df_output for more features.
        :param output_df: output df - can set a mapping.
        :param use_template: by default, if pipeline is link to a template, the query will be too. set to false to disable forcing it.
        """
        if request is None:
            raise ValueError('please provide a request.')

        if use_template and self.template_id is not None:
            request.select_template(template_id=self.template_id)

        step = PipelineStep()
        step.step_type = StepType.QUERY
        step.config = request

        if output_df is None:
            if df_name is None:
                raise ValueError('please name the output')
            step.outputs.append(PipelineIO(dataframe=df_name))
        else:
            if output_df.dataframe is None:
                raise ValueError('please set a dataframe name on output')
            step.outputs.append(output_df)

        self.steps.append(step)

    def add_transformation(self,
                           script,
                           inputs: list = None,
                           outputs: list = None,
                           input_df_names: list = None,
                           output_df_names: list = None):
        """
        add a transformation script
        :param script: name, Script or ScriptConfig.
        :param inputs: list of Pipeline I/O or dict or str for dataframe input names.
        :param outputs: list of Pipeline I/O or dict or str for dataframe output names.
        :param input_df_names: deprecated support.
        :param output_df_names: deprecated support.
        """
        step = PipelineStep()
        step.step_type = StepType.SCRIPT

        # Script Config
        if script is None:
            raise ValueError('please provide a script.')
        if isinstance(script, str):
            step.config = ScriptConfig(function=script)
        elif isinstance(script, ScriptConfig):
            step.config = script
        elif isinstance(script, Script):
            name = script.name
            if name is None:
                raise ValueError('please fetch your script or set a function name directly')
            step.config = ScriptConfig(function=name)
        else:
            raise TypeError(f'unsupported type of script {script.__class__.__name__}')

        # Inputs / Outputs

        # deprecated support - to remove 10.6/11.0
        if input_df_names is not None:
            inputs = input_df_names
        if output_df_names is not None:
            outputs = output_df_names

        if inputs is not None:
            step.inputs = PipelineIO.from_list(inputs)
        if outputs is not None:
            step.outputs = PipelineIO.from_list(outputs)

        # append
        self.steps.append(step)

    def add_model(self, config: MLModelConfig, input_df, output_df=None):
        """
        add a model step
        :param config: model configuration to define how pipeline should train and use your model.
        :param input_df: str, dict or Pipeline I/O defining input dataframe properties
        :param output_df: str, dict or Pipeline I/O defining input dataframe properties
        """
        step = PipelineStep()
        step.step_type = StepType.MODEL

        if config is None:
            raise ValueError('please provide a config.')
        if not isinstance(config, MLModelConfig):
            raise ValueError('please provide a config.')
        step.config = config

        if input_df is None:
            raise ValueError('please provide a input_df str, dict or PipelineIO')
        step.inputs.append(PipelineIO.from_obj(input_df))

        if output_df is not None:
            step.outputs.append(PipelineIO.from_obj(output_df))

        if step.config.model_key is None:
            if self.key is None:
                raise ValueError("please set a model_key in config or a pipeline key")
            step.config.model_key = self.key

        self.steps.append(step)

    def add_writer(self,
                   config: WriteConfig,
                   input_df):
        """
        add a writer step
        :param config: writer configuration to define how pipeline should write data into platform.
        :param input_df: str, dict or Pipeline I/O defining input dataframe properties
        """

        if config is None:
            raise ValueError('please provide a config.')
        if input_df is None:
            raise ValueError('please provide a input_df name.')

        step = PipelineStep()
        step.step_type = StepType.WRITE

        if input_df is None:
            raise ValueError('please provide a input_df str, dict or PipelineIO')
        step.inputs.append(PipelineIO.from_obj(input_df))

        step.config = config
        self.steps.append(step)

    def add_plot(self,
                 script,
                 df_name: str = None,
                 input_df: PipelineIO = None):
        """
        add a writer step
        :param script: script configuration to define how pipeline should execute the plot script.
        :param df_name: str deprecated usage
        :param input_df: str, dict or Pipeline I/O defining input dataframe properties
        """

        step = PipelineStep()
        step.step_type = StepType.PLOT

        if script is None:
            raise ValueError('please provide a script.')
        if isinstance(script, str):
            step.config = ScriptConfig(function=script)
        elif isinstance(script, ScriptConfig):
            step.config = script
        elif isinstance(script, Script):
            name = script.name
            if name is None:
                raise ValueError('please fetch your script or set a function name directly')
            step.config = ScriptConfig(function=name)
        else:
            raise TypeError(f'unsupported type of script {script.__class__.__name__}')

        if input_df is None:
            if df_name is None:
                raise ValueError('please name the output')
            step.inputs.append(PipelineIO(dataframe=df_name))
        else:
            if input_df.dataframe is None:
                raise ValueError('please set a dataframe name on output')
            step.inputs.append(input_df)

        self.steps.append(step)

    def check_variables(self):
        """
        verify that variables dict is a valid { "name" : "VarType" } dictionary.
        """
        if self.variables is None:
            self.variables = {}
        elif not isinstance(self.variables, dict):
            raise TypeError(f'variables must be empty nor a valid dictionary')
        for key in self.variables:
            VarType(self.variables[key])

    def api_id(self) -> str:
        """
        Id of the pipeline

        :return: string formatted UUID of the Pipeline.
        """
        return str(self.pipeline_id).upper()

    def endpoint(self) -> str:
        """
        Name of the endpoints used to manipulate pipeline.
        :return: Endpoint name.
        """
        return "Pipelines"

    def from_json(self, obj):
        """
        load from JSON dictionary representation
        """
        if "id" in obj.keys():
            self.pipeline_id = uuid.UUID(obj["id"])
        if "key" in obj.keys():
            self.key = str(obj['key'])
        if "id" not in obj.keys() and "key" not in obj.keys():
            raise KeyError("at least id or key must be set on a pipeline")
        if "templateId" in obj.keys() and obj["templateId"] is not None:
            self.template_id = uuid.UUID(obj["templateId"])
        if "experimentKey" in obj.keys() and obj["experimentKey"] is not None:
            self.experiment_id = uuid.UUID(obj["experimentKey"])
        if "variables" in obj.keys():
            if isinstance(obj["variables"], str):
                self.variables = json.loads(obj["variables"])
            else:
                self.variables = obj["variables"]
            if self.variables is not None or not isinstance(self.variables, dict):
                if isinstance(self.variables, list) and len(self.variables) == 0:
                    self.variables = {}
            self.check_variables()
        if "steps" in obj.keys():
            if isinstance(obj["steps"], str):
                steps = json.loads(obj["steps"])
            else:
                steps = obj["steps"]
            for obj_step in steps:
                step = PipelineStep()
                step.from_json(obj_step)
                self.steps.append(step)
        if "plots" in obj.keys():
            if isinstance(obj["plots"], str):
                self.plots = json.loads(obj["plots"])
            else:
                self.plots = obj["plots"]
        if "createdById" in obj.keys() and obj["createdById"] is not None:
            self.createdById = obj["createdById"]
        if "createdDate" in obj.keys() and obj["createdDate"] is not None:
            self.createdDate = obj["createdDate"]
        if "updatedById" in obj.keys() and obj["updatedById"] is not None:
            self.updatedById = obj["updatedById"]
        if "updatedDate" in obj.keys() and obj["updatedDate"] is not None:
            self.updatedDate = obj["updatedDate"]

    def to_json(self, target: str = None):
        """
        Convert to a json version of Execution definition.
        By default, use DS API format.
        """
        obj = {
            "id": str(self.pipeline_id)
        }
        if self.key is not None:
            obj["key"] = self.key
        if self.steps is not None:
            obj_steps = []
            step: PipelineStep
            for step in self.steps:
                obj_steps.append(step.to_json())
            obj["steps"] = json.dumps(obj_steps)
        if self.plots is not None:
            obj["plots"] = json.dumps(self.plots)
        if self.variables is not None:
            self.check_variables()
            obj["variables"] = json.dumps(self.variables)
        if self.template_id is not None:
            obj["templateId"] = str(self.template_id)
        if self.experiment_id is not None:
            obj["experimentKey"] = str(self.experiment_id)
        if self.createdById is not None:
            obj["createdById"] = str(self.createdById)
        if self.createdDate is not None:
            obj["createdDate"] = str(self.createdDate)
        if self.updatedById is not None:
            obj["updatedById"] = str(self.updatedById)
        if self.updatedDate is not None:
            obj["updatedDate"] = str(self.updatedDate)
        return obj

    def _load_related_entities(self, obj):
        """
        load related entities from full backend payload.
        :param obj: obj to load.
        """
        if "twins" in obj and len(obj["twins"]) == 1:
            twin_data = obj["twins"][0]
            twin_objects = {}
            twin_id = None
            if "twin" in twin_data and twin_data["twin"] is not None:
                twin_objects["twin"] = Twin()
                twin_objects["twin"].from_json(twin_data["twin"])
                twin_id = twin_objects["twin"].twin_id
            if "twinRegistration" in twin_data and twin_data["twinRegistration"] is not None:
                twin_objects["registration"] = TwinRegistration()
                twin_objects["registration"].from_json(twin_data["twinRegistration"])
                twin_id = twin_objects["registration"].twin_id
            if twin_id is not None:
                self.twins[str(twin_id)] = twin_objects
        if "template" in obj:
            self.template = Template()
            self.template.from_json(obj["template"])

    def get_model_configs(self) -> list[MLModelConfig]:
        """
        get all model configs present in this pipeline.
        :return: list of MLModelConfig
        """
        configs = []
        step: wizata_dsapi.PipelineStep
        for step in self.steps:
            if step.step_type == wizata_dsapi.StepType.MODEL:
                config: wizata_dsapi.MLModelConfig = step.config
                configs.append(config)
        return configs



import pandas
import wizata_dsapi
import numpy as np
import pandas as pd
import plotly.graph_objects as go
import plotly.express as px
import sklearn.metrics


def check_single_column_and_target_feat(context: wizata_dsapi.Context):
    """
    check_single_column_and_target_feat
    :param context:
    :return:
    """
    input_io = context.step.get_unique_input()
    if input_io.dataframe not in context.dataframes or input_io.dataframe + '.reference' not in context.dataframes:
        raise ValueError(f'impossible to find {input_io.dataframe} dataframe and reference inside the context ')

    predict_df = context.dataframes[input_io.dataframe]
    ref_df = context.dataframes[input_io.dataframe + ".reference"]
    if ref_df.ndim != 1:
        raise ValueError('please use a model that predict only one serie/dimension.')

    if "output_columns_names" not in context.properties:
        raise RuntimeError('there is no output columns in properties, r squared cannot find results')

    column_name = context.properties["output_columns_names"]
    if not isinstance(column_name, str):
        if isinstance(column_name, list) and len(column_name) == 1:
            column_name = column_name[0]
        elif isinstance(column_name, list):
            raise ValueError('please use a model that predict only one serie/dimension - mulitple column names')
        else:
            raise TypeError(f'column_name is not a str or a list {column_name.__class__.__name__}')

    if not isinstance(predict_df, pandas.DataFrame):
        raise TypeError(f'predicted dataframe is not a dataframe {predict_df.__class__.__name__}')
    predict_df = predict_df.copy()
    predict_df = predict_df[[column_name]]

    if isinstance(ref_df, pandas.Series):
        ref_df = pandas.DataFrame(ref_df, index=predict_df.index)
    return predict_df, ref_df


def confusion_matrix(context: wizata_dsapi.Context):
    """
    confusion_matrix
    :param context:
    :return:
    """
    predict_df, ref_df = check_single_column_and_target_feat(context)
    cm = sklearn.metrics.confusion_matrix(ref_df, predict_df)
    inverted_cm = np.flip(cm, axis=1)

    fig = go.Figure(data=go.Heatmap(
        z=inverted_cm,
        x=['Positive', 'Negative'],
        y=['Negative', 'Positive'],
        colorscale='RdBu',
        colorbar=dict(title='Count')
    ))

    for i in range(len(inverted_cm)):
        for j in range(len(inverted_cm[i])):
            fig.add_annotation(
                x=j,
                y=i,
                text=str(inverted_cm[i][j]),
                showarrow=False,
                font=dict(color='black' if inverted_cm[i][j] < np.max(inverted_cm) / 2 else 'white')
            )

    fig.update_layout(
        xaxis=dict(title='Predicted'),
        yaxis=dict(title='Actual')
    )

    context.set_plot(
        figure=fig,
        name="confusion_matrix"
    )


def r_squared(context: wizata_dsapi.Context):
    """
    In statistics, the coefficient of determination, denoted R2 or r2 and pronounced "R squared", is the proportion of
    the variation in the dependent variable that is predictable from the independent variable(s).
    """
    predict_df, ref_df = check_single_column_and_target_feat(context)

    r_squared = sklearn.metrics.r2_score(ref_df, predict_df)

    fig = go.Figure()
    fig.add_trace(go.Scatter(x=ref_df.values.flatten(), y=predict_df.values.flatten(),
                             mode='markers', name='Data Points'))

    min_value = min(ref_df.values.min(), predict_df.values.min())
    max_value = max(ref_df.values.max(), predict_df.values.max())
    fig.add_trace(go.Scatter(x=[min_value, max_value], y=[min_value, max_value],
                             mode='lines', line=dict(color='black', dash='dash'), showlegend=False))

    fig.add_annotation(
        x=min_value,
        y=max_value,
        text=f'R^2 = {r_squared:.4f}',
        showarrow=False,
        font=dict(color='black', size=12),
        bgcolor='lightgrey',
        bordercolor='black',
        borderwidth=1,
        borderpad=4,
        opacity=0.8
    )

    fig.update_layout(
        xaxis=dict(title='Actual'),
        yaxis=dict(title='Predicted')
    )

    context.set_plot(
        figure=fig,
        name="r_squared"
    )


def ts_chart(context: wizata_dsapi.Context):
    """
    ts_chart
    :param context:
    :return:
    """
    df = context.dataframe
    traces = []
    for column in df.columns:
        trace = go.Scatter(x=df.index, y=df[column], mode='lines', name=column)
        traces.append(trace)

    fig = go.Figure(traces)
    context.set_plot(
        figure=fig,
        name="ts_chart"
    )


def anomalies_chart(context: wizata_dsapi.Context):
    """
    anomalies_chart
    :param context:
    :return:
    """
    df = context.dataframe

    # Add Signals
    traces = []
    for column in df.columns:
        if column != "anomalies_type":
            trace = go.Scatter(x=df.index, y=df[column], mode='lines', name=column)
            traces.append(trace)
    fig = go.Figure(traces)

    # Add Anomalies as Highlighted
    anomalies_list = context.dataframe.copy()
    anomalies_list['anomaly'] = np.where(anomalies_list['anomalies_type'] != 0, 1, 0)
    anomalies_list['new_occurrence'] = np.where(
        (anomalies_list['anomaly'] != anomalies_list['anomaly'].shift(1)) |
        (anomalies_list['anomalies_type'] != anomalies_list['anomalies_type'].shift(1)), 1, 0)
    anomalies_list['new_occurrence_index'] = anomalies_list['new_occurrence'].cumsum()
    anomalies_occurrences = anomalies_list[anomalies_list['anomaly'] != 0].reset_index(). \
        groupby(['new_occurrence_index']). \
        agg({'Timestamp': ['first', 'last'], 'anomalies_type': 'first'})
    anomalies_occurrences.columns = ['from', 'to', 'anomaly_group']
    for i in anomalies_occurrences.index:
        fig.add_vrect(x0=anomalies_occurrences['from'][i], x1=anomalies_occurrences['to'][i], line_width=0,
                      fillcolor="red", opacity=0.2)

    context.set_plot(
        figure=fig,
        name="anomalies_chart"
    )


def parallel_coordinates(context: wizata_dsapi.Context):
    """
    parallel_coordinates
    :param context:
    :return:
    """
    df = context.dataframe

    fig = px.parallel_coordinates(df,
                                  color='anomalies_type',
                                  dimensions=df,
                                  color_continuous_scale=px.colors.diverging.Portland)
    context.set_plot(
        figure=fig,
        name="parallel_coordinates"
    )




import uuid
from datetime import datetime, timedelta, timezone
from .dataframe_toolkit import generate_epoch
from .api_dto import ApiDto
from .datapoint import BusinessType

filter_map = {
    '>': 'gt',
    '<': 'lt',
    '>=': 'gte',
    '<=': 'lte',
    '==': 'eq',
    '!=': 'neq'
}

class DynamicSelector:
    """
    A dynamic selector defines how datapoint(s) are fetched within a query.
        - it corresponds to a request to fetch datapoints
        - it must be used on a twin base query
    :ivar category: filter datapoints on uuid of a category or str name.
    """

    def __init__(self,
                 category = None,
                 unit = None,
                 twin_type = None,
                 agg_method = None,
                 rename = None,
                 business_type: BusinessType = None):
        self.category = category
        self.unit = unit
        self.twin_type = twin_type
        self.agg_method = agg_method
        self.rename = rename
        self.business_type = business_type
        self.datapoints = []

    @classmethod
    def from_dict(cls, obj):
        dynamic_selector = cls()
        dynamic_selector.from_json(obj)
        return dynamic_selector

    def from_json(self, obj):
        if not isinstance(obj, dict):
            raise TypeError('dynamic selector to parse is not a dict.')
        if "category" in obj:
            self.category = obj["category"]
        if "unit" in obj:
            self.unit = obj["unit"]
        if "twin_type" in obj:
            self.twin_type = obj["twin_type"]
        if "business_type" in obj:
            self.business_type = BusinessType(obj["business_type"])
        if "datapoints" in obj:
            self.datapoints = obj["datapoints"]
        if "rename" in obj:
            self.rename = obj["rename"]
        if "agg_method" in obj:
            self.agg_method = obj["agg_method"]

    def to_json(self) -> dict:
        obj = {}
        if self.category:
            obj["category"] = self.category
        if self.unit:
            obj["unit"] = self.unit
        if self.twin_type:
            obj["twin_type"] = self.twin_type
        if self.business_type:
            obj["business_type"] = self.business_type.value
        if self.datapoints:
            obj["datapoints"] = self.datapoints
        if self.agg_method:
            obj["agg_method"] = self.agg_method
        if self.rename:
            obj["rename"] = self.rename
        return obj

class RequestGroup:
    """
    Define a group statement within a structured query.
    """

    def __init__(self,
                 system_id,
                 events: list = None,
                 group_by: str = "id",
                 start_delay: int = 0,
                 end_delay: int = 0,
                 maps: list = None):
        self.system_id = system_id
        self.group_by = group_by
        self.events = events
        self.maps = maps
        self.start_delay = start_delay
        self.end_delay = end_delay

    @classmethod
    def from_dict(cls, my_dict):
        if my_dict is None or not isinstance(my_dict, dict):
            raise ValueError(f"illegal group {my_dict}")

        if "system_id" not in my_dict or my_dict["system_id"] is None:
            if not isinstance(["system_id"], str) and not isinstance(my_dict["system_id"], int):
                raise ValueError(f"please specify system_id in {my_dict}")

        group = RequestGroup(
            system_id=my_dict["system_id"]
        )

        # set group_by
        if "group_by" in my_dict:
            if my_dict["group_by"] not in ["id", "type"]:
                raise ValueError(f"illegal group {my_dict['group_by']}")
            group.group_by = my_dict["group_by"]

        # one of the two is accepted to set events
        if "event_ids" in my_dict:
            group.events = my_dict["event_ids"]
        if "events" in my_dict:
            group.events = my_dict["events"]

        if "start_delay" in my_dict:
            group.start_delay = int(my_dict["start_delay"])

        if "end_delay" in my_dict:
            group.end_delay = int(my_dict["end_delay"])

        return group


class RequestGroupMap:
    """
    Define a mapping between a set of datapoints and their corresponding event datapoint.
    :ivar event_hardware_id str: hardware id of the datapoint identifying event inside a group.
    :ivar event_datapoint: datapoint complete definition corresponding to event_hardware_id.
    :ivar datapoints_hardware_ids list: list of all related datapoints.
    :ivar datapoints dict: dict with key as hardware id and value datapoint for all mapped datapoints.
    """

    def __init__(self,
                 event_hardware_id: str,
                 datapoints_hardware_ids: list,
                 group: RequestGroup = None):
        self.event_hardware_id = event_hardware_id
        self.datapoints_hardware_ids = datapoints_hardware_ids
        self.event_datapoint = None
        self.datapoints = {}
        self.group = group


class Request(ApiDto):
    """
    request defines how to fetch data from time-series database.

    :ivar list datapoints: list of str pointing to hardware ids of datapoint or template properties name.
    :ivar datetime|str start: start range to fetch set as a datetime, a @parameter or str relative format (e.g. 'now-1d')
    :ivar datetime|str end: start range to fetch set as a datetime, a @parameter or str relative format (e.g. 'now')
    :ivar str aggregation: aggregation method ("mean", "stddev", "mode", "median", "count", "sum", "first", "last", "max", "min").
    :ivar int interval: interval passed in seconds at __init__ (stored as milliseconds).
    :ivar uuid.UUID template_id: use technical id of wizata_dsapi.Template to fetch data based on a template.
    :ivar uuid.UUID twin_id: technical id of wizata_dsapi.Twin registered on template_id.
    :ivar dict filters: filters to apply to the query pre-aggregation. nested dictionary using as first key=datapoint then value=another dictionary with key=operator (gt,lt,gte,lte,eq,neq) and value is the float value to compare too).
    :ivar dict options: dict representing query options.
    """

    @classmethod
    def route(cls):
        return "data"

    @classmethod
    def from_dict(cls, data):
        obj = Request()
        obj.from_json(data)
        return obj

    @classmethod
    def get_type(cls):
        return "pickle"

    def __init__(self,
                 datapoints=None,
                 start=None,
                 end=None,
                 agg_method='mean',
                 interval=None,
                 null='drop',
                 template_id=None,
                 template=None,
                 twin_id=None,
                 twin=None,
                 request_id=None,
                 filters=None,
                 group=None,
                 options=None,
                 field=None,
                 bucket=None,
                 tags=None):
        if request_id is None:
            request_id = uuid.uuid4()
        self.request_id = request_id
        self.function = None

        # Datapoints
        self._datapoints = []
        self.datapoints = datapoints

        # Template & Registration
        self.template = None
        self.select_template(
             template_id=template_id,
             template_key=template,
             twin_id=twin_id,
             twin_hardware_id=twin
        )

        self.start = start
        self.end = end

        self.aggregation = agg_method
        if interval is not None:
            self.interval = int(interval) / 1000
        else:
            self.interval = None

        self.filters = filters
        self.group = group
        self.options = options
        self.field = field

        self.on_off_sensor = None
        self.restart_time = None
        self.sensitivity = None
        self.dataframe = None
        self.extra_data = None
        self.target_feat = None
        self.connections = None
        self.name = None
        self.bucket = bucket
        self.tags = tags

        self.null = null

    @property
    def fields(self):
        if self.field is None:
            return ['value']
        elif isinstance(self.field, str):
            return [self.field]
        elif isinstance(self.field, list):
            return self.field
        else:
            raise TypeError(f'field must be None, str or a list of str')

    @property
    def groups(self) -> list[RequestGroup]:
        """
        return formatted groups list based on group
        :return:
        """
        if self.group is not None and self.group != {}:
            if not isinstance(self.group, list):
                return [RequestGroup.from_dict(self.group)]
            else:
                groups = []
                for group in self.group:
                    groups.append(RequestGroup.from_dict(group))
                return groups
        else:
            return []

    def get_template_id(self):
        """
        extract template id from request if present.
        :return: template id as uuid.UUID
        """
        if self.template is None:
            return

        if isinstance(self.template, str):
            try:
                template_id = uuid.UUID(self.template)
                return template_id
            except ValueError as ve:
                return
        else:
            if not isinstance(self.template, dict):
                raise ValueError(f'template in request must be a str or a dict')
            elif "template_id" not in self.template:
                return
            else:
                if isinstance(self.template['template_id'], uuid.UUID):
                    return self.template['template_id']
                else:
                    return uuid.UUID(self.template['template_id'])

    def get_template_key(self):
        """
        extract template key from request if present.
        :return: template key as str
        """
        if self.template is None:
            return

        if isinstance(self.template, str):
            return
        else:
            if not isinstance(self.template, dict):
                raise ValueError(f'template in request must be a str or a dict')
            elif "template_id" not in self.template:
                if "template_key" not in self.template:
                    return
                else:
                    return self.template["template_key"]
            else:
                return

    def get_twin_id(self):
        """
        extract twin id from request if present.
        :return: template key as uuid
        """
        if self.template is None:
            return

        if not isinstance(self.template, dict):
            return

        if "twin_id" not in self.template:
            return
        else:
            if isinstance(self.template['twin_id'], uuid.UUID):
                return self.template['twin_id']
            else:
                return uuid.UUID(self.template['twin_id'])

    def get_twin_hardware_id(self):
        """
        extract twin hardware id from request if present.
        :return: template key as str
        """
        if self.template is None:
            return

        if not isinstance(self.template, dict):
            return

        if "twin_id" not in self.template:
            if "twin_hardware_id" not in self.template:
                return
            else:
                return self.template['twin_hardware_id']
        else:
            return

    def set_twin(self, twin):
        """
        set a twin properly (str hardware Id or uuid.UUID)
        :param twin:
        :return:
        """
        if self.template is None:
            self.template = {}
        if isinstance(twin, uuid.UUID):
            self.template['twin_id'] = twin
        elif isinstance(twin, str):
            self.template['twin_hardware_id'] = twin
        else:
            raise TypeError('twin must be str or uuid on request')

    def set_template(self, template):
        """
        set a template properly (str key or uuid.UUID)
        :param template:
        :return:
        """
        if self.template is None:
            self.template = {}
        if isinstance(template, uuid.UUID):
            self.template['template_id'] = template
        elif isinstance(template, str):
            self.template['template_key'] = template
        else:
            raise TypeError('template must be str or uuid on request')

    def __format_date(self, dt_to_format):
        if isinstance(dt_to_format, datetime):
            millisec = dt_to_format.timestamp() * 1000
            return int(millisec)
        else:
            raise TypeError("date is not a valid datetime")

    def start_time(self, now=None) -> datetime:
        """
        convert a relative start time to a datetime based on a now parameter.
        :param now: override now value for relative datetime
        :return: start datetime
        """
        if self.start is None:
            raise ValueError('missing start datetime')
        elif isinstance(self.start, str):
            return datetime.fromtimestamp(generate_epoch(self.start, now=now) / 1000, timezone.utc)
        elif isinstance(self.start, datetime):
            return self.start
        else:
            raise TypeError(f'unsupported start datetime type {self.start.__class__.__name__}')

    def end_time(self, now=None) -> datetime:
        """
        convert a relative end time to a datetime based on a now parameter.
        :param now: override now value for relative datetime
        :return: end datetime
        """
        if self.end is None:
            raise ValueError('missing end datetime')
        elif isinstance(self.end, str):
            return datetime.fromtimestamp(generate_epoch(self.end, now=now) / 1000, timezone.utc)
        elif isinstance(self.end, datetime):
            return self.end
        else:
            raise TypeError(f'unsupported end datetime type {self.end.__class__.__name__}')

    @property
    def datapoints(self):
        return self._datapoints

    @datapoints.setter
    def datapoints(self, values):
        self._datapoints = []
        if values:
            if not isinstance(values, list):
                raise ValueError(f'datapoints on a query must be a list of str, dict or DynamicSelector')
            for value in values:
                if isinstance(value, str):
                    self._datapoints.append(value)
                elif isinstance(value, dict):
                    self._datapoints.append(DynamicSelector.from_dict(value))
                elif isinstance(value, DynamicSelector):
                    self._datapoints.append(value)
                else:
                    raise TypeError(f'datapoints on a query must be a list of str, dict or DynamicSelector '
                                    f'but encountered {type(value)} as {value}')

    @property
    def datapoints_without_selectors(self):
        datapoints = []
        for datapoint in self._datapoints:
            if not isinstance(datapoint, DynamicSelector):
                datapoints.append(datapoint)
        return datapoints

    @property
    def datapoints_only_selectors(self):
        datapoints = []
        for datapoint in self._datapoints:
            if isinstance(datapoint, DynamicSelector):
                datapoints.append(datapoint)
        return datapoints

    def set_aggregation(self, method, interval=None):
        """
        specifies aggregation properties
        :param method: "mean", "stddev", "mode", "median", "count", "sum", "first", "last", "max" or "min"
        :param interval: interval in ms (will be stored in seconds) can be null
        """
        if method not in self.list_agg_methods():
            raise KeyError(f'unsupported agg_method {method}.')
        self.aggregation = method
        if interval is not None:
            self.interval = int(interval) / 1000

    def list_agg_methods(self) -> list:
        """
        get a list of all authorized methods.
        :return: list with "mean", "stddev", "mode", "median", "count", "sum", "first", "last", "max" or "min"
        """
        return [
            "mean", "stddev", "mode", "median", "count", "sum", "first", "last", "max", "min", None
        ]

    def select_template(self,
                        template_id=None,
                        template_key=None,
                        twin_id=None,
                        twin_hardware_id=None):
        """
        select a template and its registration.
        :param template_id: template UUID
        :param template_key: template key ( ignored if template_id specified )
        :param twin_id: Digital Twin UUID
        :param twin_hardware_id: hardware ID of Digital Twin ( ignored if twin_id specified )
        """
        if template_id is None and template_key is None and twin_id is None and twin_hardware_id is None:
            self.template = None
            return
        else:
            self.template = {}

            template = None
            if template_key is not None:
                template = template_key
            elif template_id is not None:
                if isinstance(template_id, uuid.UUID):
                    template = template_id
                else:
                    template = uuid.UUID(template_id)
            if template is not None:
                self.set_template(template=template)

            twin = None
            if twin_hardware_id is not None:
                twin = twin_hardware_id
            elif twin_id is not None:
                if isinstance(twin_id, uuid.UUID):
                    twin = twin_id
                else:
                    twin = uuid.UUID(twin_id)
            if twin is not None:
                self.set_twin(twin=twin)

    def get_params(self):
        """
        get a list of all parameters.
        :return: list of parameters
        """
        params = []

        if self.start is not None and isinstance(self.start, str) and self.start.startswith("@"):
            params.append(self.start[1:])

        if self.end is not None and isinstance(self.end, str) and self.end.startswith("@"):
            params.append(self.end[1:])

        return list(set(params))

    def set_param(self, name: str, value):
        """
        set value of parameter based on his name.
        """
        assigned = False

        if value is None:
            raise ValueError(f'please provide a valid param value for {name}')

        if self.start is not None and isinstance(self.start, str) \
                and self.start.startswith("@") and self.start[1:] == name:
            self.start = value
            assigned = True

        if self.end is not None and isinstance(self.end, str) \
                and self.end.startswith("@") and self.end[1:] == name:
            self.end = value
            assigned = True

        if not assigned:
            raise KeyError(f'parameter {name} not found in request.')

    def to_json(self, target: str = None):
        query = {}

        if self.request_id is not None:
            query["id"] = str(self.request_id)

        datapoints = []
        for datapoint in self.datapoints:
            if isinstance(datapoint, str):
                datapoints.append(datapoint)
            elif isinstance(datapoint, dict):
                datapoints.append(datapoint)
            elif isinstance(datapoint, DynamicSelector):
                datapoints.append(datapoint.to_json())
            else:
                raise TypeError(f'datapoints on a query must be a list of str, dict or DynamicSelector '
                                f'but encountered {type(datapoint)} as {datapoint}')
        query["datapoints"] = datapoints

        if self.start is not None and self.end is not None:

            if isinstance(self.start, str):
                start = self.start
            else:
                start = self.__format_date(self.start)

            if isinstance(self.end, str):
                end = self.end
            else:
                end = self.__format_date(self.end)

            query["timeframe"] = {
                "start": start,
                "end": end
            }
        else:
            if self.group is None:
                raise KeyError("missing in query start and end date, "
                               "please use datatime format or try with a group system")
        query["aggregations"] = {
            "agg_method": self.aggregation
        }
        if self.interval:
            query["aggregations"]["interval"] = self.interval * 1000
        if self.null is not None and self.null != 'drop':
            query['null'] = self.null
        if self.template is not None:
            query['template'] = self.template
            if isinstance(self.template, dict) and 'template_id' in query['template']:
                query['template']['template_id'] = str(query['template']['template_id'])
        if self.filters is not None:
            query['filters'] = self.filters
        if self.group is not None:
            query['group'] = self.group
        if self.options is not None:
            query['options'] = self.options
        if self.field is not None:
            query['field'] = self.field
        if self.bucket is not None:
            query['bucket'] = self.bucket
        if self.tags is not None:
            query['tags'] = self.tags

        if self.target_feat is not None:
            query["target_feat"] = {
                "sensor": self.target_feat["sensor"],
                "operator": self.target_feat["operator"],
                "threshold": self.target_feat["threshold"]
            }
        if self.on_off_sensor is not None and self.restart_time is not None:
            query["restart_filter"] = {
                "on_off_sensor": self.on_off_sensor,
                "stop_restart_time": self.restart_time
            }

        if self.sensitivity is not None:
            query["sensitivity"] = self.sensitivity

        if self.extra_data is not None:
            query["extra_data"] = self.extra_data

        return query

    def from_json(self, json_data):
        if "id" in json_data.keys():
            self.request_id = uuid.UUID(json_data["id"])

        if "name" in json_data.keys():
            self.name = json_data["name"]

        datapoints = []
        if "equipments_list" in json_data.keys():
            for equipment in json_data["equipments_list"]:
                if "datapoints" not in equipment.keys():
                    raise KeyError("no 'datapoints' have been provided for equipment with id '" +
                                   str(equipment["id"]) + "'")
                datapoints.extend(equipment["datapoints"])

        if "datapoints" in json_data.keys():
            datapoints.extend(json_data["datapoints"])

        if "template" in json_data.keys():
            self.template = json_data["template"]

        if len(datapoints) == 0 and self.template is None:
            raise KeyError('at least one datapoint or template is required within a request')

        if len(datapoints) >= 0:
            self.datapoints = datapoints
        else:
            self._datapoints = []

        if "timeframe" in json_data.keys():
            if "start" not in json_data["timeframe"].keys():
                raise KeyError("No 'start time' have been selected, please set it up and re-try.")

            if isinstance(json_data["timeframe"]["start"], str):
                self.start = json_data["timeframe"]["start"]
            else:
                self.start = datetime.fromtimestamp(json_data["timeframe"]["start"] / 1000, timezone.utc)

            if "end" not in json_data["timeframe"].keys():
                raise KeyError("No 'end time' have been selected, please set it up and re-try.")

            if isinstance(json_data["timeframe"]["end"], str):
                self.end = json_data["timeframe"]["end"]
            else:
                self.end = datetime.fromtimestamp(json_data["timeframe"]["end"] / 1000, timezone.utc)
        else:
            if "group" not in json_data.keys():
                raise KeyError("No 'timeframe' and no 'group' found in object, please set a time or group selection.")

        if "aggregations" not in json_data.keys():
            raise KeyError("No 'aggregations' have been selected, please set it up and re-try.")

        if "agg_method" not in json_data["aggregations"].keys():
            self.aggregation = None
        elif json_data["aggregations"]["agg_method"] not in self.list_agg_methods():
            raise KeyError(f'unsupported agg_method {json_data["aggregations"]["agg_method"]}.')
        else:
            self.aggregation = json_data["aggregations"]["agg_method"]

        if "interval" in json_data["aggregations"].keys():
            self.interval = int(json_data["aggregations"]["interval"] / 1000)

        if "filters" in json_data.keys():
            filters = {}
            for filter_data in json_data["filters"]:
                filter_reformat = {}
                for operator in json_data["filters"][filter_data]:
                    if operator in filter_map.keys():
                        filter_reformat[filter_map[operator]] = json_data["filters"][filter_data][operator]
                    elif operator not in filter_map.values():
                        raise ValueError(f'invalid request filter operator {operator}')
                    else:
                        filter_reformat[operator] = json_data["filters"][filter_data][operator]
                filters[filter_data] = filter_reformat
            self.filters = filters
        else:
            self.filters = {}

        if "group" in json_data.keys():
            self.group = json_data["group"]
        else:
            self.group = {}

        if "options" in json_data.keys():
            self.options = json_data["options"]
        else:
            self.options = {}

        if "field" in json_data.keys():
            self.field = json_data["field"]

        if "bucket" in json_data.keys():
            self.bucket = json_data["bucket"]

        if "tags" in json_data.keys():
            self.tags = json_data["tags"]

        if "connections" in json_data.keys():
            self.connections = json_data["connections"]

        if "null" in json_data.keys():
            self.null = json_data["null"]

        if "target_feat" in json_data.keys():
            self.target_feat = json_data["target_feat"]
            if "sensor" not in self.target_feat.keys():
                raise KeyError("No 'sensor' have been declared inside the target feature, this is a technical error.")
            if "operator" not in self.target_feat.keys():
                raise KeyError("No 'operator' have been declared inside the target feature, this is a technical error.")
            if "threshold" not in self.target_feat.keys():
                raise KeyError("No 'threshold' have been declared inside the target feature, this is a technical error.")

        if "restart_filter" in json_data.keys():
            self.on_off_sensor = json_data["restart_filter"]["on_off_sensor"]
            self.restart_time = json_data["restart_filter"]["stop_restart_time"]

        if "sensitivity" in json_data.keys():
            self.sensitivity = json_data["sensitivity"]

        if "extra_data" in json_data.keys():
            self.extra_data = json_data["extra_data"]





from typing import List, Iterator, Union, Optional
import os
from .api_dto import ApiDto, ApiDtoInterface
from datetime import datetime, timezone


def get_bool(obj, name: str):
    if isinstance(obj[name], str) and obj[name].lower() == "false":
        return False
    else:
        return bool(obj[name])


class ModelIdentifierInfo:
    """
    define metadata associated to an identifier (regardless of its alias)
    """

    def __init__(self,
                 identifier: str = None,
                 active_alias: str = None):
        self.identifier = identifier
        self.active_alias = active_alias

    def to_json(self):
        obj = {}
        if self.active_alias is not None:
            obj["active_alias"] = self.active_alias
        return obj

    def from_json(self, json_dict):
        if "active_alias" in json_dict:
            self.active_alias = json_dict["active_alias"]

class ModelFile:
    """
    define a model file.

    :ivar path str: file name relative path e.g. trained_model.pkl
    :ivar content bytes[]: array of content bytes
    """

    def __init__(self,
                 full_path: str = None,
                 path: str = None,
                 name: str = None,
                 extension: str = None,
                 last_modified: datetime = None,
                 size: int = None,
                 content = None):
        self.full_path = full_path
        self.name = name
        self.extension = extension
        self.path = path
        self.last_modified = last_modified
        self.size = size
        self.content = content


    @property
    def path(self) -> Optional[str]:
        if self.name and self.extension:
            return f"{self.name}.{self.extension}"
        return self.name or None

    @path.setter
    def path(self, value: Optional[str]):
        if not value:
            self.name = None
            self.extension = None
            return

        base = os.path.basename(value.strip())
        name_part, ext = os.path.splitext(base)
        self.name = name_part or None
        self.extension = ext.lstrip(".").lower() if ext else None

    def to_json(self):
        """
        convert this entity in a dict that can be json serializable
        :return: dict
        """
        obj = {}
        if self.full_path is not None:
            obj["full_path"] = str(self.full_path)
        if self.name is not None:
            obj["name"] = self.name
        if self.extension is not None:
            obj["extension"] = self.extension
        if self.path is not None:
            obj["path"] = self.path
        if self.last_modified is not None:
            obj["last_modified"] = self.last_modified.timestamp() * 1000.0
        if self.size is not None:
            obj["size"] = self.size
        return obj

    def from_json(self, obj):
        """
        load this entity from a dict
        """
        if "full_path" in obj.keys():
            self.full_path = obj["full_path"]
        if "name" in obj.keys():
            self.name = obj["name"]
        if "extension" in obj.keys():
            self.extension = obj["extension"]
        if not self.name and not self.extension and "path" in obj:
            self.path = obj["path"]
        if "last_modified" in obj.keys():
            self.last_modified = datetime.fromtimestamp(obj["last_modified"] / 1000, tz=timezone.utc)
        if "size" in obj.keys():
            self.size = obj["size"]


class ModelInfo:
    """
    define a pointer to a machine learning model.

    :ivar model_type str: mlflow model flavour
    :ivar files list: list of ModelFile of model content.
    """

    def __init__(self,
                 key: str,
                 twin_hardware_id: str = None,
                 property_value: str = None,
                 alias: str = None,
                 model_type: str = None,
                 file_format: str = 'pkl',
                 source: str = 'wizata',
                 property_name: str = None,
                 trained_model = None,
                 scaler = None,
                 files = None,
                 updated_date = None
                 ):
        # information identifying a model
        self.key = key
        self.twin_hardware_id = twin_hardware_id
        self.property_value = property_value
        self.alias = alias
        self.file_format = file_format
        if self.file_format not in ["pt", "pkl"]:
            raise ValueError("file_format must be 'pt' or 'pkl'")
        self.source = source
        self.model_type = model_type
        if files is None:
            self.files = []
        else:
            self.files = files
            for file in files:
                if not isinstance(file, ModelFile):
                    raise TypeError("file is not a ModelFile with files")
        self.is_active = False
        self.updated_date = None

        # files attached to model when loaded
        self.trained_model = trained_model
        self.scaler = scaler

        # temporary properties during model generation not generally stored
        self.property_name = property_name
        self.input_columns = None
        self.has_target_feat = False
        self.label_counts = 0

        # api
        self._api = None

    def bind_api(self, api:ApiDtoInterface):
        """
        internal method to bind the api to the dto.
        :param api: api client
        :return: None
        """
        self._api = api

    @classmethod
    def split_identifier(cls, identifier: str):
        """
        split the identifier into four parts
        :param identifier: identifier.
        :return: key, twin, property, alias
        """

        # extract alias
        if "@" in identifier:
            split_identifier_for_alias = identifier.split("@")
            identifier = split_identifier_for_alias[0]
            alias = split_identifier_for_alias[1]
        else:
            alias = None

        # extract key, twin, property
        identifiers = identifier.split('.')
        key = identifiers[0]
        twin_hardware_id = None
        if len(identifiers) > 1:
            twin_hardware_id = identifiers[1]
        property_value = None
        if len(identifiers) > 2:
            property_value = identifiers[2]

        return key, twin_hardware_id, property_value, alias

    def identifier(self, include_alias: bool = False) -> str:
        """
        returns the complete string identifier for this model.
        :param include_alias: include the alias pointer or leave it to target the default version.
        :return: complete identifier of a model.
        """
        if self.key is None:
            raise KeyError('please specific a model key')
        identifier = self.key

        if self.twin_hardware_id is not None:
            identifier += f".{self.twin_hardware_id}"

        if self.property_value is not None:
            identifier += f".{self.property_value}"

        if include_alias and self.alias is not None:
            identifier += f"@{self.alias}"

        return identifier

    def add_file(self, file: ModelFile):
        """
        add a path to list of known path.
        detect depending on file type further actions.
        :param file: ModelFile
        :return: None
        """
        if file.name == "trained_model":
            self.file_format = file.extension
        self.files.append(file)

    def to_json(self) -> dict:
        """
        convert this entity in a dict that can be json serializable
        :return: dict
        """
        obj = {
            "key": self.key,
            "file_format": self.file_format,
            "source": self.source
        }
        if self.twin_hardware_id is not None:
            obj["twin_hardware_id"] = str(self.twin_hardware_id)
        if self.property_value is not None:
            obj["property_value"] = self.property_value
        if self.alias is not None:
            obj["alias"] = self.alias
        if self.model_type is not None:
            obj["model_type"] = self.model_type
        if self.property_name is not None:
            obj["property_name"] = self.property_name
        if self.is_active is not None:
            obj["is_active"] = self.is_active
        if self.identifier is not None:
            obj["identifier"] = self.identifier(include_alias=True)
        if len(self.files) > 0:
            obj["files"] = [file.to_json() for file in self.files]
        if self.updated_date is not None:
            obj["updatedDate"] = str(self.updated_date)
        return obj

    def from_json(self, obj):
        """
        load this entity from a dict
        """
        if "key" in obj.keys():
            self.key = obj["key"]
        if "twin_hardware_id" in obj.keys():
            self.twin_hardware_id = obj["twin_hardware_id"]
        if "property_value" in obj.keys():
            self.property_name = obj["property_value"]
        if "alias" in obj.keys():
            self.alias = obj["alias"]
        if "model_type" in obj.keys():
            self.model_type = obj["model_type"]
        if "file_format" in obj.keys():
            self.file_format = obj["file_format"]
        if "source" in obj.keys():
            self.source = obj["source"]
        if "property_name" in obj.keys():
            self.property_name = obj["property_name"]
        if "is_active" in obj.keys():
            self.is_active = get_bool(obj, name="is_active")
        if "updatedDate" in obj.keys() and obj["updatedDate"] is not None:
            self.updated_date = obj["updatedDate"]
        if "files" in obj.keys():
            for obj_file in obj["files"]:
                model_file = ModelFile()
                model_file.from_json(obj_file)
                self.add_file(model_file)

    def load(self):
        """
        load the trained model from the repository.
        """
        if self._api is None:
            raise RuntimeError("api is not bound to the dto use bind_api()")
        self._api.load_model(self)


class ModelList:
    """
    used to conveniently manipulate a list of models.
    """

    def __init__(self):
        self.models: List[ModelInfo] = []
        self.identifiers_info = {}

    def __iter__(self) -> Iterator[ModelInfo]:
        return iter(self.models)

    def __len__(self) -> int:
        return len(self.models)

    def exists(self, model: ModelInfo) -> bool:
        """
        check if a model is within the model list.
        :param model: identifier to check.
        :return: True if it exists.
        """
        return any(model_in_list.identifier(include_alias=True) == model.identifier(include_alias=True) for model_in_list in self.models)

    def __getitem__(self, key: Union[int, str, ModelInfo]) -> ModelInfo:
        """
        find a model within list based on index, identifier or ModelInfo.
        :param key: identifier or ModelInfo or index
        :return: the model_info
        """
        if isinstance(key, int):
            return self.models[key]

        elif isinstance(key, str):
            if "@" not in key:
                return self.select_active_model(identifier=key)
            for model in self.models:
                if model.identifier(include_alias=True) == key:
                    return model
            raise KeyError(f"model with identifier '{key}' not found within this ModelList.")

        elif isinstance(key, ModelInfo):
            identifier = key.identifier(include_alias=True)
            for model in self.models:
                if model.identifier(include_alias=True) == identifier:
                    return model
            raise KeyError(f"model with identifier '{identifier}' not found within this ModelList.")

        else:
            raise TypeError("ModelList indices must be int, str or ModelInfo.")

    def select_active_model(self, identifier: str) -> ModelInfo:
        """
        return the active model based on active status or latest one if none active.
        :param identifier: identifier
        :return: active model
        """
        models = []
        for model in self.models:
            if model.identifier(include_alias=False) == identifier:
                if model.is_active:
                    return model
                else:
                    models.append(model)
        return max(models, key=lambda f: f.updated_date, default=None)

    def append(self, model: ModelInfo):
        self.models.append(model)

    def _process_info(self):
        """
        apply all info within the model list (setting model active, ...) to keep data coherence
        :return:
        """
        for identifier in self.identifiers_info:
            if self.identifiers_info[identifier].active_alias is not None:
                for model in self.models:
                    if model.identifier(include_alias=False) == identifier and model.alias == self.identifiers_info[identifier].active_alias:
                        model.is_active = True
                    elif model.identifier(include_alias=False) == identifier:
                        model.is_active = False


class MLModelConfig(ApiDto):
    """
    a model config defines execution properties within a pipeline.
    usually to define how a pipeline should train and predict with your model.

    :ivar format: by default 'pkl' for pickle, accept 'pt' for PyTorch model using tensors (none = 'pkl')
    :ivar bool by_twin: define if pipeline need to train and use different model by twin item.
    :ivar bool by_property: define if pipeline need to train and use different model based on 'property_name'.
    :ivar list features: datapoint list to refine columns if necessary.
    :ivar str function: name of the function used instead of 'predict' on model inference.
    :ivar str model_key: key of the model to store (or use story property if dynamic).
    :ivar str model_type: reserved for 'mlflow' defines model flavor such as 'sklearn' or 'pyfunc'.
    :ivar str model_alias: reserved for 'mlflow' set the alias to target and use inside a pipeline.
    :ivar str model_format: type of model (by default 'pkl') accept also 'pt' for PyTorch model using tensors (none = 'pkl')
    :ivar dict properties_mapping: dict to map properties expected by the script with the properties from the context.
    :ivar str property_name: define which property is looked for when using 'by_property'.
    :ivar str source: define name of model storage to use 'wizata'.
    :ivar str|list target_feat: target feature name(s) to remove from df in 'predict' mode.
    :ivar str train_script: name of function referencing the script to train the model.
    :ivar float train_test_split_pct: percentage repartition to split the data for training and scoring.
    :ivar str train_test_split_type: type of splitting desired to split the train dataframe.
    :ivar bool output_append: true - default - append output to input dataframe. false to retrieve only 'predict' column(s).
    :ivar list output_columns_names: name list to rename columns inside output dataframe.
    :ivar str output_prefix: set a prefix to put before output column names (default or custom).
    """

    def __init__(self,
                 train_script=None,
                 train_test_split_pct: float = 1.0,
                 train_test_split_type: str = "ignore",
                 function: str = "predict",
                 properties_mapping=None,
                 target_feat=None,
                 output_append: bool = True,
                 output_columns_names: list = None,
                 output_prefix: str = None,
                 features: list = None,
                 model_key: str = None,
                 model_type: str = None,
                 model_alias: str = None,
                 model_format: str = 'pkl',
                 by_twin: bool = False,
                 by_property: bool = False,
                 property_name: str = None,
                 source: str = "wizata"
                 ):

        # training
        self.train_script = train_script
        self.train_test_split_pct = train_test_split_pct
        self.train_test_split_type = train_test_split_type
        self.function = function
        self.properties_mapping = properties_mapping

        # features management
        self.target_feat = target_feat
        self.features = features
        self.output_columns_names = output_columns_names
        self.output_append = output_append
        self.output_prefix = output_prefix

        # key and identification
        self.model_key = model_key
        self.by_twin = by_twin
        self.by_property = by_property
        self.property_name = property_name
        self.model_format = model_format

        # source
        self.source = source
        self.model_type = model_type
        self.model_alias = model_alias

    def create_model_info(self,
                          hardware_id: str = None,
                          property_value: str = None) -> ModelInfo:
        """
        create model info corresponding to the configuration.
        :param hardware_id: provide a hardware id for this model if by_twin.
        :param property_value: provide a value for this model if by_property.
        :return:
        """
        if self.by_twin and hardware_id is None:
            raise ValueError('hardware_id is required if by_twin to create a model info')
        if self.by_property and property_value is None:
            raise ValueError('property_value is required if by_property to create a model info')
        model_info = ModelInfo(
            key=self.model_key,
            twin_hardware_id=hardware_id,
            property_value=property_value,
            source=self.source,
            alias=self.model_alias,
            file_format=self.model_format,
            model_type=self.model_type,
            property_name=self.property_name
        )
        return model_info

    def from_json(self, obj):

        # Managed deprecated fields
        if "model_key_type" in obj.keys() and obj["model_key_type"] is not None:
            if obj["model_key_type"] == "template":
                self.by_twin = True
            elif obj["model_key_type"] == "variable":
                self.by_property = True
                if "model_key" not in obj.keys() or obj["model_key"] is None:
                    raise KeyError('model_key must be declared in the config.')
                self.property_name = obj["model_key"]

        # training info
        if "train_script" in obj.keys() and obj["train_script"] is not None:
            self.train_script = obj["train_script"]
        if "train_test_split_pct" in obj.keys() and obj["train_test_split_pct"] is not None:
            self.train_test_split_pct = float(obj["train_test_split_pct"])
        if "train_test_split_type" in obj.keys() and obj["train_test_split_type"] is not None:
            self.train_test_split_type = obj["train_test_split_type"]
        if "function" in obj.keys() and obj["function"] is not None:
            self.function = obj["function"]
        if "properties_mapping" in obj:
            self.properties_mapping = obj["properties_mapping"]

        # features info
        if "target_feat" in obj.keys() and obj["target_feat"] is not None:
            if isinstance(obj["target_feat"], str) or isinstance(obj["target_feat"], list):
                self.target_feat = obj["target_feat"]
            else:
                raise ValueError(f'target_feat should be a str or a list with columns name to remove.')

        if "features" in obj.keys() and obj["features"] is not None:
            self.features = obj["features"]

        # output management
        if "output_append" in obj.keys():
            self.output_append = get_bool(obj, name="output_append")

        if "output_prefix" in obj.keys() and obj["output_prefix"] is not None:
            self.output_prefix = obj["output_prefix"]
        if "output_columns_names" in obj.keys() and obj["output_columns_names"] is not None:
            if isinstance(obj["output_columns_names"], str):
                self.output_columns_names = [obj["output_columns_names"]]
            elif isinstance(obj["output_columns_names"], list):
                self.output_columns_names = obj["output_columns_names"]
            else:
                raise ValueError(f'output_property or output_columns_names should be a list or a string')
        elif "output_property" in obj.keys() and obj["output_property"] is not None \
                and obj["output_property"] != "result":
            if isinstance(obj["output_property"], str):
                self.output_columns_names = [obj["output_property"]]
            elif isinstance(obj["output_property"], list):
                self.output_columns_names = obj["output_property"]
            else:
                raise ValueError(f'output_property or output_columns_names should be a list or a string')

        # source
        if "source" in obj.keys() and obj["source"] is not None:
            self.source = obj["source"]
            if self.source not in ["wizata", "mlflow"]:
                raise ValueError("source must be wizata or mlflow")

        # key and target
        if "model_key" not in obj.keys() or obj["model_key"] is None:
            raise KeyError('model_key must be declared in the config.')
        self.model_key = obj["model_key"]
        if "by_twin" in obj.keys():
            self.by_twin = get_bool(obj, name="by_twin")
        if "by_property" in obj.keys():
            self.by_property = get_bool(obj, name="by_property")
        if "property_name" in obj.keys() and obj["property_name"] is not None:
            self.property_name = obj["property_name"]
        if "model_type" in obj.keys() and obj["model_type"] is not None:
            self.model_type = obj["model_type"]
        if "model_alias" in obj.keys() and obj["model_alias"] is not None:
            self.model_alias = obj["model_alias"]
        if "model_format" in obj.keys() and obj["model_format"] is not None:
            self.model_format = obj["model_format"]

    def to_json(self, target: str = None):
        obj = {
            "source": self.source
        }

        # training info
        if self.train_script is not None:
            obj["train_script"] = str(self.train_script)
        if self.train_test_split_pct is not None:
            obj["train_test_split_pct"] = float(self.train_test_split_pct)
        if self.train_test_split_type is not None:
            obj["train_test_split_type"] = str(self.train_test_split_type)
        if self.features is not None:
            obj["features"] = self.features
        if self.properties_mapping is not None and isinstance(self.properties_mapping, dict):
            obj["properties_mapping"] = self.properties_mapping

        # features info
        if self.target_feat is not None:
            obj["target_feat"] = self.target_feat
        if self.output_append is not None:
            obj["output_append"] = str(self.output_append)
        if self.output_columns_names is not None:
            obj["output_columns_names"] = self.output_columns_names
        if self.output_prefix is not None:
            obj["output_prefix"] = self.output_prefix
        if self.function is not None:
            obj["function"] = self.function

        # key and target
        if self.model_key is not None:
            obj["model_key"] = self.model_key
        if self.model_type is not None:
            obj["model_type"] = self.model_type
        if self.model_alias is not None:
            obj["model_alias"] = self.model_alias
        if self.by_twin is not None:
            obj["by_twin"] = str(self.by_twin)
        if self.by_property is not None:
            obj["by_property"] = str(self.by_property)
        if self.property_name is not None:
            obj["property_name"] = self.property_name
        if self.model_format  is not None:
            obj["model_format"] = self.model_format

        return obj

    def has_target_feat(self) -> bool:
        """
        determine if configuration possess a target feature
        """
        if self.target_feat is None:
            return False

        if isinstance(self.target_feat, str):
            return True
        elif isinstance(self.target_feat, list):
            if len(self.target_feat) == 0:
                return False
            else:
                return True
        else:
            raise TypeError(f'unsupported target_feat type {self.target_feat.__class__.__name__}')

